//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include "vision_api_cxx_image_analysis_event_args.hpp"
#include "vision_api_cxx_image_analysis_error_reason.hpp"

#include <azac_api_cxx_details_property_collection.hpp>
#include <vision_api_cxx_common.hpp>
#include <vision_api_cxx_details_session_error_details.hpp>

namespace Azure {
namespace AI {
namespace Vision {
namespace ImageAnalysis {

using namespace Azure::AI::Vision::Core::_detail;

/// <summary>
/// A representation of an error associated with an image analysis result.
/// </summary>
class ImageAnalysisErrorDetails :
    private Core::_detail::SessionErrorDetails<ImageAnalysisErrorReason, ImageAnalysisEventArgs, ImageAnalysisErrorDetails, ImageAnalysisResult>
{
private:

    using BaseDetails = Core::_detail::SessionErrorDetails<ImageAnalysisErrorReason, ImageAnalysisEventArgs, ImageAnalysisErrorDetails, ImageAnalysisResult>;

public:
    /// <summary>
    /// Creates an object that contains additional error information for a failed ImageAnalysisResult.
    /// </summary>
    /// <remarks>
    /// ImageAnalysisErrorDetail objects can only be created from ImageAnalysisResults that stopped due to an
    /// error. This corresponds to a value of ImageAnalysisResultReason::Error when calling ImageAnalysisResult::GetReason.
    /// Attempting to create an ImageAnalysisErrorDetails from a result that did not stop due to an error will
    /// return a nullptr.
    /// </remarks>
    /// <param name="result">The result object from a failed image analysis operation</param>
    /// <returns>The newly created ImageAnalysisErrorDetails wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<ImageAnalysisErrorDetails> FromResult(std::shared_ptr<ImageAnalysisResult> result)
    {
        if (result && result->GetReason() != ImageAnalysisResultReason::Error)
        {
            return nullptr;
        }
        else
        {
            return BaseDetails::FromResult(result);
        }
    }

    /// <summary>
    /// Get the detailed error message.
    /// </summary>
    /// <returns>
    /// The error message.
    /// </returns>
    std::string GetMessage() { return BaseDetails::GetMessage(); }

    /// <summary>
    /// Get the detailed error message.
    /// </summary>
    /// <returns>
    /// The error message.
    /// </returns>
    template<class T = std::string>
    AI::Core::_detail::enable_if_w_or_string_t<T> GetMessage() { return BaseDetails::GetMessage<T>(); }

    /// <summary>
    /// Gets a standardized code for the error.
    /// </summary>
    /// <returns>
    /// The error code.
    /// </returns>
    int GetErrorCode() { return BaseDetails::GetErrorCode(); }

    /// <summary>
    /// Gets the category of the error.
    /// </summary>
    /// <returns>
    /// The category of the error.
    /// </returns>
    ImageAnalysisErrorReason GetReason() { return BaseDetails::GetReason(); }

private:
    friend class Azure::AI::Core::_detail::ProtectedAccess<ImageAnalysisErrorDetails>;

    static std::shared_ptr<ImageAnalysisErrorDetails> FromHandle(AZAC_HANDLE handle)
    {
        auto ptr = new ImageAnalysisErrorDetails(handle);
        return std::shared_ptr<ImageAnalysisErrorDetails>(ptr);
    }

    explicit ImageAnalysisErrorDetails(AZAC_HANDLE propertiesHandle)
        : BaseDetails(propertiesHandle) {}

    explicit operator AZAC_HANDLE() { return AI::Core::_detail::ProtectedAccess<ImageAnalysisErrorDetails>::HandleFromPtr(this); }

private:
    AZAC_DISABLE_DEFAULT_CTORS(ImageAnalysisErrorDetails);
};

} } } } // Azure::AI::Vision::ImageAnalysis
