//
// Copyright (c) Microsoft. All rights reserved.
// See https://aka.ms/azai/vision/license for the full license information.
//

#pragma once
#include <vision_api_cxx_source.hpp>
#include <azac_api_cxx_details_property_collection.hpp>
#include <vision_api_cxx_common.hpp>

#include <initializer_list>
#include <numeric>
#include <ostream>
#include <string>
#include <vector>
#include <map>
#include <sstream>

namespace Azure {
namespace AI {
namespace Vision {
namespace ImageAnalysis {

/// <summary>
/// Defines the supported visual features to detect in an image.
/// </summary>
enum class ImageAnalysisFeature
{
    /// <summary>
    /// Tags the image with a detailed list of recognizable objects, living beings, scenery, and actions that appear in the image.
    /// The language of the tags can be specified by setting the property ImageAnalysisOptions::SetLanguage.
    /// </summary>
    Tags = 0,

    /// <summary>
    /// Generates a human-readable phrase that describes the image content, in one of the supported languages.
    /// </summary>
    /// <remarks>
    /// Gender neutral caption can be requested by calling ImageAnalysisOptions::SetGenderNeutralCaption(true).
    /// The language can be specified by setting the property ImageAnalysisOptions::SetLanguage.
    /// </remarks>
    Caption,

    /// <summary>
    /// Dense Captions provides more details than ImageAnalysisFeature::Caption, by generating one sentence
    /// descriptions of up to 10 regions of the image in addition to describing the whole image.
    /// </summary>
    /// <remarks>
    /// Dense Captions also returns bounding box coordinates of the described image region.
    /// Gender neutral caption can be requested by calling ImageAnalysisOptions::SetGenderNeutralCaption(true).
    /// The language can be specified by setting the property ImageAnalysisOptions::SetLanguage.
    /// </remarks>
    DenseCaptions,

    /// <summary>
    /// Detects various objects within an image, including their approximate location.
    /// Object names are only available in English at the moment.
    /// </summary>
    Objects,

    /// <summary>
    /// Detects people in the image, including their approximate location.
    /// </summary>
    People,

    /// <summary>
    /// Also known as Read or OCR. Performs Optical Character Recognition (OCR)
    /// and returns the text detected in the image, including the approximate location
    /// of every text line and word.
    /// </summary>
    Text,

    /// <summary>
    /// Also known as SmartCrops. Returns recommendations for image crop operations that
    /// preserve content (for example for thumbnail generation)
    /// </summary>
    /// <remarks>
    /// Provide requested aspect ratios by calling ImageAnalysisOptions::SetCroppingAspectRatios
    /// </remarks>
    CropSuggestions
};

/// <summary>
/// Defines the segmentation mode supported by the Image Analysis service.
/// </summary>
enum class ImageSegmentationMode
{
    /// <summary>
    /// The default value. No segmentation is performed.
    /// </summary>
    None = 0,

    /// <summary>
    /// Background removal. Segmentation results in a PNG image of the detected
    /// foreground object with a transparent background.
    /// </summary>
    BackgroundRemoval,

    /// <summary>
    /// Foreground matting. Segmentation results in a grayscale alpha matte PNG image
    /// showing the opacity of the detected foreground object
    /// </summary>
    ForegroundMatting
};

/// <summary>
/// Represents the configuration options that control the function of the ImageAnalyzer
/// </summary>
/// <remarks>
/// If you are doing Image Analysis using the standard model, you must call
/// SetFeatures() and set one or more visual features to analyze. There is no
/// default selection for visual features. If you are using a custom model or doing Image Segmentation,
/// you do not need to specify visual features.
/// </remarks>
class ImageAnalysisOptions
{
public:
    /// <summary>
    /// Set a list of one or more visual features to extract from the image
    /// </summary>
    /// <remarks>
    /// If you are using the default model, at least one visual feature
    /// must be specified.
    /// If you are using a custom-trained model (by calling SetModelName), you do
    /// not need to specify any visual features, as they are implied from the model.
    /// If you are doing Image Segmentation (by calling SetSegmentationMode), you do not need to specify visual features.
    /// Note that you can extract visual features (by calling SetFeatures and/or SetModelName) or
    /// do segmentation (by calling SetSegmentationMode) but you cannot do both at the same time.
    /// </remarks>
    /// <param name="ImageAnalysisFeatures">A list of visual features</param>
    void SetFeatures(const std::vector<ImageAnalysisFeature>& ImageAnalysisFeatures)
    {
        std::map<ImageAnalysisFeature, std::string> ImageAnalysisFeaturesMap =
        {
            {ImageAnalysisFeature::Caption, "caption"},
            {ImageAnalysisFeature::DenseCaptions, "denseCaptions"},
            {ImageAnalysisFeature::Objects, "objects"},
            {ImageAnalysisFeature::Text, "read"},
            {ImageAnalysisFeature::Tags, "tags"},
            {ImageAnalysisFeature::People, "people"},
            {ImageAnalysisFeature::CropSuggestions, "smartCrops"}
        };

        std::string packedImageAnalysisFeatures = std::accumulate(ImageAnalysisFeatures.begin(), ImageAnalysisFeatures.end(), std::string(),
            [&ImageAnalysisFeaturesMap](const std::string& ss, const ImageAnalysisFeature& v)
            {
                std::string s = ImageAnalysisFeaturesMap[v];
                return ss.empty() ? s : ss + "," + s;
            });

        m_options->Set("image.analysis.option.visualfeatures", packedImageAnalysisFeatures);
    }

    /// <summary>
    /// Gets the list of one or more visual features to extract from the image.
    /// </summary>
    /// <returns>
    /// A list of visual features
    /// </returns>
    std::vector<ImageAnalysisFeature> GetFeatures()
    {
        std::vector<ImageAnalysisFeature> features;
        std::string packedFeatures =  m_options->Get("image.analysis.option.visualfeatures", "");
        if (!packedFeatures.empty())
        {
            std::map<std::string, ImageAnalysisFeature> ImageAnalysisFeaturesMap =
            {
                {"caption", ImageAnalysisFeature::Caption},
                {"denseCaptions", ImageAnalysisFeature::DenseCaptions},
                {"objects", ImageAnalysisFeature::Objects},
                {"read", ImageAnalysisFeature::Text},
                {"tags", ImageAnalysisFeature::Tags},
                {"people", ImageAnalysisFeature::People},
                {"smartCrops", ImageAnalysisFeature::CropSuggestions}
            };
            std::stringstream ss(packedFeatures);
            std::string item;
            while (std::getline(ss, item, ','))
            {
                ImageAnalysisFeature feature = ImageAnalysisFeaturesMap[item];
                features.push_back(feature);
            }
        }
        return features;
    }

    /// <summary>
    /// Sets the list of aspect ratios that crop suggestions should attempt to fit (also known as smart crops)
    /// </summary>
    /// <remarks>
    /// An aspect ratio is calculated by dividing the target crop width by the height.
    ///
    /// Supported values are between 0.75 and 1.8 (inclusive).
    ///
    /// Calling this method is only relevant when ImageAnalysisFeature::CropSuggestions is set as 
    /// one of the image features to analyze (see ImageAnalysisOptions::SetFeatures method).
    ///
    /// If SetCroppingAspectRatios() is not called, but ImageAnalysisFeature::CropSuggestions is specified
    /// as a feature, the service will return one crop suggestion with an aspect ratio it sees fit between
    /// 0.5 and 2.0 (inclusive).
    /// </remarks>
    /// <param name="aspectRatios">A list of aspect rations (a positive number)</param>
    void SetCroppingAspectRatios(const std::vector<double>& aspectRatios)
    {
        std::string packedAspectRatios = std::accumulate(aspectRatios.begin(), aspectRatios.end(), std::string(),
            [](const std::string& ss, const double& aspectRatio)
            {
                std::string s = std::to_string(aspectRatio);
                return ss.empty() ? s : ss + "," + s;
            });
        m_options->Set("image.analysis.option.croppingaspectratios", packedAspectRatios);
    }

    /// <summary>
    /// Get the list of aspect ratios to be used for cropping (also known as smart crops)
    /// </summary>
    /// <returns>
    /// A list of aspect ratios. This list will be empty if aspect ratios have not been set before</param>
    /// </returns>
    /// <remarks>
    /// An aspect ratio is calculated by dividing the target crop width by the height.
    ///
    /// Supported values are between 0.75 and 1.8 (inclusive).
    /// </remarks>
    std::vector<double> GetCroppingAspectRatios()
    {
        std::vector<double> aspectRatios;
        std::string packedAspectRatios = m_options->Get("image.analysis.option.croppingaspectratios", "");
        if (!packedAspectRatios.empty())
        {
            std::stringstream ss(packedAspectRatios);
            std::string item;
            while (std::getline(ss, item, ','))
            {
                aspectRatios.push_back(std::stod(item));
            }
        }
        return aspectRatios;
    }

    /// <summary>
    /// Sets the language that Image Analysis should use in the results.
    /// </summary>
    /// <remarks>
    /// This language should be provided as an ISO 639-1 code, e.g. "en" for English or "fr" for French.
    /// If this value is not set, the default value is "en" for English.
    /// See https://aka.ms/cv-languages for a list of supported language codes and which
    /// visual features are supported for each language.
    /// </remarks>
    /// <param name="language">The language code</param>
    void SetLanguage(const std::string& language)
    {
        m_options->Set("image.analysis.option.language", language);
    }

    /// <summary>
    /// Gets the language that Image Analysis should use in the results.
    /// </summary>
    /// <returns>
    /// The language the service will use in Image Analysis results.
    /// </returns>
    /// <remarks>
    /// The language is in format ISO 639-1, e.g. "en" for English or "fr" for French.
    /// If this value is not set, the default value is "en" for English.
    /// See https://aka.ms/cv-languages for a list of supported language codes and which
    /// visual features are supported for each language.
    /// </remarks>
    std::string GetLanguage()
    {
        return m_options->Get("image.analysis.option.language", "en");
    }

    /// <summary>
    /// Sets the model version that Image Analysis Service should use.
    /// If this option is not set, the default is \"latest\".
    /// </summary>
    /// <remarks>
    /// \"latest\" is the only value currently supported by the service.
    /// In future service updates, supported model versions will be \"latest\" or
    /// in the form \"YYYY-MM-DD\" or \"YYYY-MM-DD-preview\",
    /// where YYYY, MM, DD are year, month, and day respectively.
    /// </remarks>
    /// <param name="modelVersion">The model version</param>
    void SetModelVersion(const std::string& modelVersion)
    {
        m_options->Set("image.analysis.option.modelversion", modelVersion);
    }

    /// <summary>
    /// Gets the model version that Image Analysis Service should use.
    /// If this option was not set, the default is \"latest\".
    /// </summary>
    /// <returns>
    /// The AI model version used by the service. The only value
    /// supported at the moment is \"latest\".
    /// </returns>
    std::string GetModelVersion()
    {
        return m_options->Get("image.analysis.option.modelversion", "latest");
    }

    /// <summary>
    /// Sets the gender neutrality of the Image Analysis caption result.
    /// </summary>
    /// <param name="genderNeutral">If \"true\", the caption will not have gendered terms.
    /// If \"true\", the words \"Man/Woman\" will be replaced by \"Person\", and \"Boy/Girl\" will be replaced by \"Child\".
    /// If not set, defaults to \"false\."</param>
    /// <remarks>
    /// Only relevant if ImageAnalysisFeature::Caption is included in the Features list (See ImageAnalysisOptions::SetFeatures).
    /// </remarks>
    void SetGenderNeutralCaption(const bool genderNeutral)
    {
        m_options->Set("image.analysis.option.genderneutralcaption", genderNeutral ? "true" : "false");
    }

    /// <summary>
    /// Gets the gender neutrality of the Image Analysis caption result.
    /// </summary>
    /// <returns>
    /// \"true\" if the returned caption will not have gendered terms. \"false\" (the default) otherwise.
    /// </returns>
    bool GetGenderNeutralCaption()
    {
        return (m_options->Get("image.analysis.option.genderneutralcaption", "false") == "true");
    }

    /// <summary>
    /// Sets the name of the custom-trained model that the Image Analysis Service should use.
    /// If this option is not set, the default (standard) model will be used.
    /// Note that you can extract visual features (by calling SetFeatures and/or SetModelName) or
    /// do segmentation (by calling SetSegmentationMode) but you cannot do both at the same time.
    /// </summary>
    /// <param name="modelName">The custom model name</param>
    void SetModelName(const std::string& modelName)
    {
        m_options->Set("image.analysis.option.modelname", modelName);
    }

    /// <summary>
    /// Gets the name of the custom-trained model that the Image Analysis Service uses.
    /// </summary>
    /// <returns>
    /// The custom model name, or an empty string if a custom model
    /// has not been set (in which case the standard model will be used).
    /// </returns>
    std::string GetModelName()
    {
        return m_options->Get("image.analysis.option.modelname", "");
    }

    /// <summary>
    /// Sets the segmentation mode that the Image Analysis Service should use.
    /// </summary>
    /// <remarks>
    /// By setting either ImageSegmentationMode::BackgroundRemoval or ImageSegmentationMode::ForegroundMatting,
    /// the Image Analysis service will perform a segmentation operation, and if succesfull,
    /// will return a single PNG image of the resulting segmentation.
    /// By default no segmentation is done.
    /// Note that you can extract visual features (by calling SetFeatures and/or SetModelName) or
    /// do segmentation (by calling SetSegmentationMode) but you cannot do both at the same time.
    /// </remarks>
    /// <param name="segmentationMode">The desired segmentation mode</param>
    void SetSegmentationMode(const ImageSegmentationMode segmentationMode)
    {
        std::map<ImageSegmentationMode, std::string> segmentationModeMap =
        {
            {ImageSegmentationMode::None, ""},
            {ImageSegmentationMode::BackgroundRemoval, "backgroundRemoval"},
            {ImageSegmentationMode::ForegroundMatting, "foregroundMatting"}
        };

        m_options->Set("image.analysis.option.segmentationmode", segmentationModeMap[segmentationMode]);
    }

    /// <summary>
    /// Gets the current segmentation mode.
    /// </summary>
    /// <returns>
    /// The current segmentation mode.
    /// </returns>
    ImageSegmentationMode GetSegmentationMode()
    {
        std::map<std::string, ImageSegmentationMode> segmentationModeMap =
        {
            {"", ImageSegmentationMode::None},
            {"backgroundRemoval", ImageSegmentationMode::BackgroundRemoval},
            {"foregroundMatting", ImageSegmentationMode::ForegroundMatting}
        };

        return segmentationModeMap[m_options->Get("image.analysis.option.segmentationmode", "")];
    }

    /// <summary>
    /// Represents the advanced options and parameters used in an Image Analysis operation.
    /// </summary>
    class AdvancedOptions
    {
    private:
        friend class ImageAnalysisOptions;
        template<typename Target> using ProtectedAccess = Azure::AI::Core::_detail::ProtectedAccess<Target>;
        using PropertyCollection = AI::Core::_detail::PropertyCollection<>;
        std::shared_ptr<PropertyCollection> m_options;
        std::shared_ptr<Input::VisionSource> m_referenceObject;

    public:
        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new string value</param>
        void Set(const std::string& option, const char* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Gets the string value for the specified option.
        /// </summary>
        /// <param name="option">The Image Analysis option</param>
        /// <param name="defaultValue">The default value to return if the option was not set</param>
        /// <returns>
        /// The string value of the specified option
        /// </returns>
        std::string Get(const std::string& option, const std::string& defaultValue)
        {
            return m_options->Get(option, defaultValue);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new string value</param>
        void Set(const std::wstring&& option, const wchar_t* value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new string value</param>
        void Set(const std::string& option, const std::string& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the string value for the specified option.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new string value</param>
        void Set(const std::wstring&& option, const std::wstring&& value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property boolean value for the specified key value.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new boolean value</param>
        void Set(const std::string& option, bool value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property boolean value for the specified key value.
        /// </summary>
        /// <param name="option">The option key</param>
        /// <param name="value">The new boolean value</param>
        void Set(const std::wstring&& option, bool value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property integer value for the specified key value.
        /// </summary>
        /// <param name="option">The integer value's key</param>
        /// <param name="value">The new integer value</param>
        void Set(const std::string& option, int value)
        {
            m_options->Set(option, value);
        }

        /// <summary>
        /// Sets the property integer value for the specified key value.
        /// </summary>
        /// <param name="option">The integer value's key</param>
        /// <param name="value">The new integer value</param>
        void Set(const std::wstring&& option, int value)
        {
            m_options->Set(option, value);
        }

    private:

        friend class Azure::AI::Core::_detail::ProtectedAccess<AdvancedOptions>;

        static std::shared_ptr<AdvancedOptions> Create()
        {
            auto ptr = new AdvancedOptions();
            return std::shared_ptr<AdvancedOptions>(ptr);
        }

        explicit AdvancedOptions() :
            m_options(PropertyCollection::Create())
        {
        }

        explicit operator AZAC_HANDLE()
        {
            AZAC_HANDLE hOptions = ProtectedAccess<PropertyCollection>::HandleFromPtr(m_options.get());
            if (m_referenceObject)
            {   // Copy options to property collection within vision source of reference object and pass it back.
                AZAC_HANDLE hReferenceObjectProperties = ProtectedAccess<Input::VisionSourceProperties>::HandleFromPtr(&m_referenceObject->Properties);
                ai_core_properties_copy(hOptions, hReferenceObjectProperties, nullptr);
                return hReferenceObjectProperties;
            }
            return hOptions;
        }

        AZAC_DISABLE_COPY_AND_MOVE(AdvancedOptions);
    };

private:

    template<typename Target> using ProtectedAccess = Azure::AI::Core::_detail::ProtectedAccess<Target>;

    std::shared_ptr<AdvancedOptions> m_options;

public:

    /// <summary>
    /// Initializes a new instance of the ImageAnalysisOptions class.
    /// </summary>
    /// <returns>The newly created ImageAnalysisOptions wrapped inside a std::shared_ptr</returns>
    static std::shared_ptr<ImageAnalysisOptions> Create()
    {
        auto ptr = new ImageAnalysisOptions();
        return std::shared_ptr<ImageAnalysisOptions>(ptr);
    }

    /// <summary>
    /// Destructs an instance of the ImageAnalysisOptions class.
    /// </summary>
    ~ImageAnalysisOptions() = default;

    /// <summary>
    /// Advanced options and parameters
    /// </summary>
    AdvancedOptions& Advanced;

private:

    friend class Azure::AI::Core::_detail::ProtectedAccess<ImageAnalysisOptions>;

    explicit ImageAnalysisOptions() :
        m_options(ProtectedAccess<AdvancedOptions>::Create()),
        Advanced(*m_options.get())
    {
    }

    explicit operator AZAC_HANDLE()
    {
        return ProtectedAccess<AdvancedOptions>::HandleFromPtr(m_options.get());
    }

    AZAC_DISABLE_COPY_AND_MOVE(ImageAnalysisOptions);
};

} } } } // Azure::AI::Vision::ImageAnalysis
